# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import itertools
import logging
from math import prod
from typing import FrozenSet, Iterator, List, Optional, Sequence, Tuple, Union

import numpy as np

import cuequivariance as cue
from cuequivariance.etc.linalg import (
    basis_intersection,
    perm_compose,
    perm_inverse,
    perm_sign,
    round_to_sqrt_rational,
    sparsify_matrix,
)
from cuequivariance.group_theory import Irrep, irreps_array
from cuequivariance.group_theory.irreps_array.irrep_utils import into_list_of_irrep

logger = logging.getLogger(__name__)


# The code below is adapted from https://github.com/e3nn/e3nn-jax/blob/245e17eb23deaccad9f2c9cfd40fe40515e3c074/e3nn_jax/_src/reduced_tensor_product.py

"""
History of the different versions of the code:
- Initially developed by Mario Geiger in `e3nn`
- Ported in julia by Song Kim https://github.com/songk42/ReducedTensorProduct.jl
- Ported in `e3nn-jax` by Mario Geiger
- Optimized the symmetric case by Ameya Daigavane and Mario Geiger
- imported and modified in cuequivariance by NVIDIA Corporation & Affiliates
"""


###### Frontend functions ######


def reduced_tensor_product_basis(
    formula_or_irreps_list: Union[str, List[irreps_array.Irreps]],
    *,
    layout: irreps_array.IrrepsLayout = irreps_array.mul_ir,
    epsilon: float = 1e-5,
    keep_ir: Optional[Union[irreps_array.Irreps, List[Irrep]]] = None,
    _use_optimized_implementation: bool = True,
    **irreps_dict,
) -> irreps_array.NumpyIrrepsArray:
    r"""Reduce a tensor product of multiple irreps subject to some permutation symmetry given by a formula.

    Args:
        formula_or_irreps_list (str or list of Irreps): a formula of the form ``ijk=jik=ikj`` or ``ijk=-jki``.
            The left hand side is the original formula and the right hand side are the signed permutations.
            If no index symmetry is present, a list of irreps can be given instead.

        epsilon (float): the tolerance for the Gram-Schmidt orthogonalization. Default: ``1e-5``
        keep_ir (list of Irrep): irrep to keep in the output. Default: keep all irrep
        irreps_dict (dict): the irreps of each index of the formula. For instance ``i="1x1o"``.

    Returns:
        IrrepsArray: The change of basis
            The shape is ``(d1, ..., dn, irreps_out.dim)``
            where ``di`` is the dimension of the index ``i`` and ``n`` is the number of indices in the formula.
    """
    if not isinstance(layout, irreps_array.IrrepsLayout):
        layout = irreps_array.IrrepsLayout[layout]

    if isinstance(formula_or_irreps_list, (tuple, list)):
        irreps_tuple = tuple(formula_or_irreps_list)
        assert all(isinstance(irreps, irreps_array.Irreps) for irreps in irreps_tuple)
        perm_repr: FrozenSet[Tuple[int, Tuple[int, ...]]] = frozenset(
            {(1, tuple(range(len(irreps_tuple))))}
        )
        if keep_ir is not None:
            keep_ir = frozenset(
                into_list_of_irrep(irreps_tuple[0].irrep_class, keep_ir)
            )
        return _entrypoint(
            irreps_tuple,
            perm_repr,
            keep_ir,
            layout,
            epsilon,
            _use_optimized_implementation,
        )

    formula = formula_or_irreps_list
    f0, perm_repr = germinate_perm_repr(formula)

    # irreps_dict = {i: irreps_array.Irreps(irs) for i, irs in irreps_dict.items()}
    assert all(isinstance(irs, irreps_array.Irreps) for irs in irreps_dict.values())

    for i in irreps_dict:
        if len(i) != 1:
            raise TypeError(f"got an unexpected keyword argument '{i}'")

    for _sign, p in perm_repr:
        f = "".join(f0[i] for i in p)
        for i, j in zip(f0, f):
            if (
                i in irreps_dict
                and j in irreps_dict
                and irreps_dict[i] != irreps_dict[j]
            ):
                raise RuntimeError(f"irreps of {i} and {j} should be the same")
            if i in irreps_dict:
                irreps_dict[j] = irreps_dict[i]
            if j in irreps_dict:
                irreps_dict[i] = irreps_dict[j]

    for i in f0:
        if i not in irreps_dict:
            raise RuntimeError(f"index {i} has no irreps associated to it")

    for i in irreps_dict:
        if i not in f0:
            raise RuntimeError(
                f"index {i} has an irreps but does not appear in the fomula"
            )

    irreps_tuple = tuple(irreps_dict[i] for i in f0)

    if keep_ir is not None:
        keep_ir = frozenset(into_list_of_irrep(irreps_tuple[0].irrep_class, keep_ir))
    return _entrypoint(
        irreps_tuple, perm_repr, keep_ir, layout, epsilon, _use_optimized_implementation
    )


def _symmetric_perm_repr(n: int):
    return frozenset((1, p) for p in itertools.permutations(range(n)))


def reduced_symmetric_tensor_product_basis(
    irreps: irreps_array.Irreps,
    degree: int,
    *,
    layout: irreps_array.IrrepsLayout = irreps_array.mul_ir,
    epsilon: float = 1e-5,
    keep_ir: Optional[Union[irreps_array.Irreps, List[Irrep]]] = None,
    _use_optimized_implementation: bool = True,
) -> irreps_array.NumpyIrrepsArray:
    r"""Reduce a symmetric tensor product, usually called for a single irrep.

    Args:
        irreps (Irreps): the irreps of each index.
        degree (int): the degree of the tensor product. i.e. the number of indices.
        epsilon (float): the tolerance for the Gram-Schmidt orthogonalization. Default: ``1e-5``
        keep_ir (list of Irrep): irrep to keep in the output. Default: keep all irrep

    Returns:
        IrrepsArray: The change of basis
            The shape is ``(d, ..., d, irreps_out.dim)``
            where ``d`` is the dimension of ``irreps``.
    """
    assert isinstance(irreps, irreps_array.Irreps)

    if not isinstance(layout, irreps_array.IrrepsLayout):
        layout = irreps_array.IrrepsLayout[layout]

    if keep_ir is not None:
        from .irrep_utils import into_list_of_irrep

        keep_ir = frozenset(into_list_of_irrep(irreps.irrep_class, keep_ir))

    perm_repr: FrozenSet[Tuple[int, Tuple[int, ...]]] = _symmetric_perm_repr(degree)
    return _entrypoint(
        tuple([irreps] * degree),
        perm_repr,
        keep_ir,
        layout,
        epsilon,
        _use_optimized_implementation,
    )


def _antisymmetric_perm_repr(n: int):
    return frozenset((perm_sign(p), p) for p in itertools.permutations(range(n)))


def reduced_antisymmetric_tensor_product_basis(
    irreps: irreps_array.Irreps,
    degree: int,
    *,
    layout: irreps_array.IrrepsLayout = irreps_array.mul_ir,
    epsilon: float = 1e-5,
    keep_ir: Optional[Union[irreps_array.Irreps, List[Irrep]]] = None,
    _use_optimized_implementation: bool = True,
) -> irreps_array.NumpyIrrepsArray:
    r"""Reduce an antisymmetric tensor product.

    Args:
        irreps (Irreps): the irreps of each index.
        degree (int): the degree of the tensor product. i.e. the number of indices.
        epsilon (float): the tolerance for the Gram-Schmidt orthogonalization. Default: ``1e-5``
        keep_ir (list of Irrep): irrep to keep in the output. Default: keep all irrep

    Returns:
        IrrepsArray: The change of basis
            The shape is ``(d, ..., d, irreps_out.dim)``
            where ``d`` is the dimension of ``irreps``.
    """
    assert isinstance(irreps, irreps_array.Irreps)

    if not isinstance(layout, irreps_array.IrrepsLayout):
        layout = irreps_array.IrrepsLayout[layout]

    if keep_ir is not None:
        from .irrep_utils import into_list_of_irrep

        keep_ir = frozenset(into_list_of_irrep(irreps.irrep_class, keep_ir))

    perm_repr: FrozenSet[Tuple[int, Tuple[int, ...]]] = _antisymmetric_perm_repr(degree)
    return _entrypoint(
        tuple([irreps] * degree),
        perm_repr,
        keep_ir,
        layout,
        epsilon,
        _use_optimized_implementation,
    )


###### Backend functions ######


def _entrypoint(
    irreps_tuple: Tuple[irreps_array.Irreps, ...],
    perm_repr: FrozenSet[Tuple[int, Tuple[int, ...]]],
    keep_ir: Optional[FrozenSet[Irrep]],
    layout: irreps_array.IrrepsLayout,
    epsilon: float,
    _use_optimized_implementation: bool,
) -> irreps_array.NumpyIrrepsArray:
    output = _main_cached_recursive(
        irreps_tuple,
        perm_repr,
        keep_ir,
        epsilon,
        _use_optimized_implementation,
    )

    # Sparsifying only at the end seems to give better sparsification
    output = _sparsify(output)

    if layout == cue.ir_mul:
        x = output.array

        for axis, irreps in enumerate(irreps_tuple + (output.irreps,)):
            x = np.moveaxis(x, axis, -1)
            x = (
                irreps_array.NumpyIrrepsArray(irreps, x, irreps_array.mul_ir)
                .change_layout(cue.ir_mul)
                .array
            )
            x = np.moveaxis(x, -1, axis)

        return irreps_array.NumpyIrrepsArray(output.irreps, x, cue.ir_mul)
    return output


@functools.lru_cache(maxsize=None)
def _main_cached_recursive(
    irreps_tuple: Tuple[irreps_array.Irreps, ...],
    perm_repr: FrozenSet[Tuple[int, Tuple[int, ...]]],
    keep_ir: Optional[FrozenSet[Irrep]],
    epsilon: float,
    _use_optimized_implementation: bool,
) -> irreps_array.NumpyIrrepsArray:
    degree = len(irreps_tuple)

    # Optimized case
    if (
        _use_optimized_implementation
        and perm_repr == _symmetric_perm_repr(degree)
        and degree > 1
    ):
        irreps = irreps_tuple[0]

        if irreps.num_irreps > 1:
            logger.debug(
                f"_reduced_tensor_product_basis {irreps_tuple}, {perm_repr}, keep_ir={keep_ir}, use _optimized_reduced_symmetric_tensor_product_basis"
            )
            return _rounding(
                _optimized_reduced_symmetric_tensor_product_basis(
                    irreps, degree, epsilon=epsilon, keep_ir=keep_ir
                )
            )

    logger.debug(
        f"_reduced_tensor_product_basis {irreps_tuple}, {perm_repr}, keep_ir={keep_ir}"
    )

    # General case
    dims = tuple(irreps.dim for irreps in irreps_tuple)

    bases = [
        (
            frozenset({i}),
            irreps_array.NumpyIrrepsArray(
                irreps,
                np.reshape(
                    np.eye(irreps.dim),
                    (1,) * i
                    + (irreps.dim,)
                    + (1,) * (len(irreps_tuple) - i - 1)
                    + (irreps.dim,),
                ),
                irreps_array.mul_ir,
            ),
        )
        for i, irreps in enumerate(irreps_tuple)
    ]

    while True:
        if len(bases) == 0:
            raise RuntimeError("Tensor Product raised to the 0th power is not defined.")

        if len(bases) == 1:
            f, b = bases[0]
            assert f == frozenset(range(len(irreps_tuple)))
            if keep_ir is not None:
                b = b.filter(keep=keep_ir)
            return b.regroup()

        if len(bases) == 2:
            (fa, a) = bases[0]
            (fb, b) = bases[1]
            f = frozenset(fa | fb)
            ab = reduce_basis_product(a, b, keep_ir)
            if len(subrepr_permutation(f, perm_repr)) == 1:
                return ab.regroup()
            p = reduce_subgroup_permutation(f, perm_repr, dims)
            ab = constrain_rotation_basis_by_permutation_basis(ab, p, epsilon=epsilon)
            return _rounding(ab.regroup())

        # greedy algorithm
        min_p = np.inf
        best = None

        for i in range(len(bases)):
            for j in range(i + 1, len(bases)):
                (fa, _) = bases[i]
                (fb, _) = bases[j]
                f = frozenset(fa | fb)
                p_dim = reduce_subgroup_permutation(f, perm_repr, dims, return_dim=True)
                if p_dim < min_p:
                    min_p = p_dim
                    best = (i, j, f)

        i, j, f = best
        del bases[j]
        del bases[i]
        del i, j
        sub_irreps = tuple(irreps_tuple[i] for i in f)
        sub_perm_repr = subrepr_permutation(f, perm_repr)
        keep = keep_ir
        for i, irreps in enumerate(irreps_tuple):
            if i not in f:
                keep = cue.selection_rule_product(keep, {ir for _, ir in irreps})
        ab = _main_cached_recursive(
            sub_irreps,
            sub_perm_repr,
            keep,
            epsilon,
            _use_optimized_implementation,
        )
        ab = ab.reshape(
            tuple(dims[i] if i in f else 1 for i in range(len(dims))) + (-1,)
        )
        bases = [(f, ab)] + bases


###### Special case for symmetric tensor product ######


def generate_tuples_with_fixed_sum(length: int, sum: int):
    """Generates all non-negative integer tuples of a certain length with the specified sum.

    Examples:
        generate_tuples_with_fixed_sum(1, 3) = (3,)
        generate_tuples_with_fixed_sum(2, 3) = (3, 0) (2, 1) (1, 2) (0, 3)
    """
    if length == 1:
        yield (sum,)
        return

    if sum == 0:
        yield tuple(0 for _ in range(length))
        return

    for first in range(sum, -1, -1):
        for subtuple in generate_tuples_with_fixed_sum(length - 1, sum - first):
            yield (first,) + subtuple


def repeat_indices(indices: Sequence[int], powers: Sequence[int]) -> List[int]:
    """Given [i1, i2, ...] and [p1, p2, ...], returns [i1, i1, ... (p1 times), i2, i2, ... (p2 times), ...]"""
    repeated_indices = []
    for index, power in zip(indices, powers):
        repeated_indices.extend([index] * power)
    return repeated_indices


def generate_permutations(
    seq: Sequence[float],
) -> Iterator[Tuple[Sequence[float], Sequence[int]]]:
    """Generates permutations of a sequence along with the indices used to create the permutation."""
    indices = range(len(seq))
    for permuted_indices in itertools.permutations(indices):
        permuted_sequence = tuple(seq[index] for index in permuted_indices)
        yield permuted_sequence, permuted_indices


def cumsum_before(seq: Sequence[float]) -> np.ndarray:
    """Returns the cumulative sum before every index.

    For example, cumsum_before([1, 2, 3]) == [0, 1, 3].
    """
    return np.cumsum([0, *seq])[:-1]


def cumsum_after(seq: Sequence[float]) -> np.ndarray:
    """Returns the cumulative sum after every index.

    For example, cumsum_after([1, 2, 3]) == [5, 3, 0].
    """
    return cumsum_before(seq[::-1])[::-1]


def reshape_for_basis_product(
    terms: Sequence[irreps_array.NumpyIrrepsArray], non_zero_powers: Sequence[float]
):
    """Adds extra axes to each term to be compatible for reduce_basis_product()."""
    term_powers_cumsum_before = cumsum_before(non_zero_powers)
    term_powers_cumsum_after = cumsum_after(non_zero_powers)

    def reshape_term_for_basis_product(
        index: int, term: irreps_array.NumpyIrrepsArray
    ) -> irreps_array.NumpyIrrepsArray:
        new_shape = (
            (1,) * term_powers_cumsum_before[index]
            + term.shape[:-1]
            + (1,) * term_powers_cumsum_after[index]
            + term.shape[-1:]
        )
        return term.reshape(new_shape)

    return [
        reshape_term_for_basis_product(index, term) for index, term in enumerate(terms)
    ]


def _optimized_reduced_symmetric_tensor_product_basis(
    irreps: irreps_array.Irreps,
    degree: int,
    *,
    epsilon: float,
    keep_ir: Optional[List[Irrep]] = None,
):
    r"""Reduce a symmetric tensor product.

    Args:
        irreps (Irreps): the irreps of each index.
        degree (int): the degree of the tensor product. i.e. the number of indices.
        epsilon (float): the tolerance for the Gram-Schmidt orthogonalization. Default: ``1e-5``
        keep_ir (list of Irrep): irrep to keep in the output. Default: keep all irrep

    Returns:
        IrrepsArray: The change of basis
            The shape is ``(irreps.dim, ..., irreps.dim, irreps_out.dim)``
    """
    assert isinstance(irreps, irreps_array.Irreps)

    def compute_padding_for_term(irrep_indices: Sequence[int]) -> List[Tuple[int, int]]:
        """Computes the padding for the given term at each index.

        This is required because the output change-of-basis must have the
        shape (irreps.dim, ..., irreps.dim, irreps_out.dim),
        which means all input axes must have the same length.

        For example, when computing ir^3 = (ir_1 + ir_2)^3,
        we get a term corresponding to (ir_1)^3 with shape (ir_1.dim, ..., ir_1.dim, ...).
        We need to pad this term to have shape (ir.dim, ..., ir.dim, ...) instead.
        A similar logic applies for all of the terms.
        """
        inp_irreps_dims = [mul * ir.dim for mul, ir in irreps]
        inp_irreps_dims_cumsum_before = cumsum_before(inp_irreps_dims)
        inp_irreps_dims_cumsum_after = cumsum_after(inp_irreps_dims)

        def compute_padding_for_irrep_index(irrep_index: int):
            dims_before = inp_irreps_dims_cumsum_before[irrep_index]
            dims_after = inp_irreps_dims_cumsum_after[irrep_index]
            return (dims_before, dims_after)

        return [
            compute_padding_for_irrep_index(irrep_index)
            for irrep_index in irrep_indices
        ] + [(0, 0)]

    irreps = irreps_array.Irreps(
        irreps.irrep_class, [(1, ir) for mul, ir in irreps for _ in range(mul)]
    )

    # Precompute powers of irreps.
    irreps_powers = {}
    for i, mul_ir in enumerate(irreps):
        irreps_powers[i] = [
            irreps_array.NumpyIrrepsArray(
                irreps_array.Irreps(irreps.irrep_class, [irreps.irrep_class.trivial()]),
                np.asarray([1.0]),
                irreps_array.mul_ir,
            )
        ]
        for n in range(1, degree + 1):
            irreps_powers[i].append(
                reduced_symmetric_tensor_product_basis(
                    irreps_array.Irreps(irreps.irrep_class, [mul_ir]),
                    n,
                    epsilon=epsilon,
                    keep_ir=cue.selection_rule_product(
                        keep_ir,
                        cue.selection_rule_power(
                            irreps.irrep_class, {ir for _, ir in irreps}, degree - n
                        ),
                    ),
                )
            )

    # Take all products of irreps whose powers sum up to degree.
    # For example, if we are computing (ir1 + ir2)^3, we would consider terms of the form:
    # - ir_1 ir_1 ir_1
    # - ir_1 ir_1 ir_2, ir_1 ir_2 ir_1, ir_2 ir_1 ir_1
    # - ir_1 ir_2 ir_2, ir_2 ir_1 ir_2, ir_2 ir_2 ir_1
    # - ir_2 ir_2 ir_2
    # where the terms on the same line will be averaged over.
    # Each line above corresponds to a unique tuple:
    # - (3, 0)
    # - (2, 1)
    # - (1, 2)
    # - (0, 3)
    # indicating the powers of the individual irreps ir_1 and ir_2.
    # Note that possible many terms correspond to the same tuple,
    # since the tuple does not indicate the degree of multiplication.
    symmetric_product = []
    for term_powers in generate_tuples_with_fixed_sum(len(irreps), degree):
        term_powers = list(term_powers)

        non_zero_indices = [i for i, n in enumerate(term_powers) if n != 0]
        non_zero_powers = [n for n in term_powers if n != 0]
        non_zero_indices_repeated = tuple(
            repeat_indices(non_zero_indices, non_zero_powers)
        )

        # Add axes to all terms, so that they have the same number of input axes.
        non_zero_terms = [
            irreps_powers[i][n] for i, n in zip(non_zero_indices, non_zero_powers)
        ]
        non_zero_terms_reshaped = reshape_for_basis_product(
            non_zero_terms, non_zero_powers
        )

        # Compute basis product, two terms at a time.
        if len(non_zero_terms_reshaped) == 1:
            product_basis = non_zero_terms_reshaped[0].filter(keep=keep_ir)
        else:
            current_term = non_zero_terms_reshaped[0]
            for next_term in non_zero_terms_reshaped[1:-1]:
                current_term = reduce_basis_product(current_term, next_term)
            last_term = non_zero_terms_reshaped[-1]
            product_basis = reduce_basis_product(
                current_term, last_term, keep_ir=keep_ir
            )

        if product_basis.irreps.dim == 0:
            continue

        shape = (irreps.dim,) * degree + (product_basis.irreps.dim,)
        sum_of_permuted_bases = np.zeros_like(product_basis.array, shape=shape)
        seen_permutations = set()

        # Now, average over the different permutations.
        for permuted_indices_repeated, permuted_axes in generate_permutations(
            non_zero_indices_repeated
        ):
            # Keep track of which permutations we have seen.
            # Don't repeat permutations!
            if permuted_indices_repeated in seen_permutations:
                continue
            seen_permutations.add(permuted_indices_repeated)

            # Permute axes according to this term.
            permuted_product_basis_array = np.transpose(
                product_basis.array, permuted_axes + (len(permuted_axes),)
            )

            # Add padding.
            padding = compute_padding_for_term(permuted_indices_repeated)
            slices = tuple(
                slice(start, total - stop)
                for (start, stop), total in zip(padding, shape)
            )

            sum_of_permuted_bases[slices] += permuted_product_basis_array

        # Normalize the sum of bases.
        symmetrized_sum_of_permuted_bases = sum_of_permuted_bases / np.sqrt(
            len(seen_permutations)
        )
        product_basis = irreps_array.NumpyIrrepsArray(
            product_basis.irreps, symmetrized_sum_of_permuted_bases, irreps_array.mul_ir
        )
        symmetric_product.append(product_basis)

    # Filter out irreps, if needed.
    basis = cue.concatenate(symmetric_product)
    basis = basis.regroup()
    return basis


###### Utilities ######


@functools.lru_cache(maxsize=None)
def germinate_perm_repr(
    formula: str,
) -> Tuple[str, FrozenSet[Tuple[int, Tuple[int, ...]]]]:
    """Convert the formula (generators) into a group."""
    formulas = [
        (-1 if f.startswith("-") else 1, f.replace("-", "")) for f in formula.split("=")
    ]
    s0, f0 = formulas[0]
    assert s0 == 1

    for _s, f in formulas:
        if len(set(f)) != len(f) or set(f) != set(f0):
            raise RuntimeError(f"{f} is not a permutation of {f0}")
        if len(f0) != len(f):
            raise RuntimeError(f"{f0} and {f} don't have the same number of indices")

    # `perm_repr` is a list of (sign, permutation of indices)
    # each formula can be viewed as a permutation of the original formula
    perm_repr = {
        (s, tuple(f.index(i) for i in f0)) for s, f in formulas
    }  # set of generators (permutations)

    # they can be composed, for instance if you have ijk=jik=ikj
    # you also have ijk=jki
    # applying all possible compositions creates an entire group
    while True:
        n = len(perm_repr)
        perm_repr = perm_repr.union([(s, perm_inverse(p)) for s, p in perm_repr])
        perm_repr = perm_repr.union(
            [
                (s1 * s2, perm_compose(p1, p2))
                for s1, p1 in perm_repr
                for s2, p2 in perm_repr
            ]
        )
        if len(perm_repr) == n:
            break  # we break when the set is stable => it is now a group \o/

    return f0, frozenset(perm_repr)


_cache_reduce_basis_product = {}


def reduce_basis_product(
    basis1: irreps_array.NumpyIrrepsArray,
    basis2: irreps_array.NumpyIrrepsArray,
    keep_ir: Optional[List[Irrep]] = None,
) -> irreps_array.NumpyIrrepsArray:
    """Reduce the product of two basis."""
    basis1 = basis1.regroup()
    basis2 = basis2.regroup()

    assert basis1.layout == basis2.layout
    layout = basis1.layout

    if keep_ir is not None:
        assert all(isinstance(ir, Irrep) for ir in keep_ir)
        keep_ir = frozenset(keep_ir)

    assert basis1.array.dtype == np.float64
    assert basis2.array.dtype == np.float64

    cache_key = (
        keep_ir,
        basis1.irreps,
        basis2.irreps,
        basis1.shape,
        basis2.shape,
        hash(basis1.array.tobytes()),
        hash(basis2.array.tobytes()),
    )
    if cache_key in _cache_reduce_basis_product:
        return _cache_reduce_basis_product[cache_key]

    new_irreps: List[Tuple[int, Irrep]] = []
    new_list = []

    for (mul1, ir1), x1 in zip(basis1.irreps, basis1.segments):
        for (mul2, ir2), x2 in zip(basis2.irreps, basis2.segments):
            for ir in ir1 * ir2:
                if keep_ir is not None and ir not in keep_ir:
                    continue

                w = ir1.clebsch_gordan(ir1, ir2, ir)
                mulcg = len(w)
                if layout == irreps_array.mul_ir:
                    x = np.einsum("...ui,...vj,wijk->...uvwk", x1, x2, w)
                elif layout == cue.ir_mul:
                    x = np.einsum("...iu,...jv,wijk->...kuvw", x1, x2, w)
                x = np.reshape(
                    x, x.shape[:-4] + layout.shape((mul1 * mul2 * mulcg, ir))
                )
                new_list.append(x)
                new_irreps.append((mul1 * mul2 * mulcg, ir))

    new_irreps = irreps_array.Irreps(basis1.irreps.irrep_class, new_irreps)
    leading_shape = np.broadcast_shapes(basis1.shape[:-1], basis2.shape[:-1])
    out = cue.from_segments(new_irreps, new_list, layout, leading_shape).regroup()
    _cache_reduce_basis_product[cache_key] = out
    return out


def constrain_rotation_basis_by_permutation_basis(
    rotation_basis: irreps_array.NumpyIrrepsArray,
    permutation_basis: np.ndarray,
    *,
    epsilon=1e-5,
    round_fn=lambda x: x,
) -> irreps_array.NumpyIrrepsArray:
    """Constrain a rotation basis by a permutation basis.

    Args:
        rotation_basis (e3nn.IrrepsArray): A rotation basis
        permutation_basis (np.ndarray): A permutation basis

    Returns:
        e3nn.IrrepsArray: A rotation basis that is constrained by the permutation basis.
    """
    assert rotation_basis.shape[:-1] == permutation_basis.shape[1:]

    layout = rotation_basis.layout
    assert layout == irreps_array.mul_ir

    perm = np.reshape(
        permutation_basis,
        (permutation_basis.shape[0], prod(permutation_basis.shape[1:])),
    )  # (free, dim)

    new_irreps: List[Tuple[int, Irrep]] = []
    new_list: List[np.ndarray] = []

    for ir in sorted({ir for mul, ir in rotation_basis.irreps}):
        idx = [i for i, (mul, ir_) in enumerate(rotation_basis.irreps) if ir == ir_]
        rot_basis = np.concatenate([rotation_basis.segments[i] for i in idx], axis=-2)
        mul = rot_basis.shape[-2]
        R = rot_basis[..., 0]
        R = np.reshape(R, (-1, mul)).T  # (mul, dim)

        # optimization:
        perm_opt = perm[~np.all(perm[:, ~np.all(R == 0, axis=0)] == 0, axis=1)]
        # NOTE: this optimization work only because perm rows don't share non-zero elements

        P, _ = basis_intersection(R, perm_opt, epsilon=epsilon, round_fn=round_fn)

        new_irreps.append((len(P), ir))
        new_list.append(round_fn(np.einsum("vu,...ui->...vi", P, rot_basis)))

    new_irreps = irreps_array.Irreps(rotation_basis.irreps.irrep_class, new_irreps)
    return cue.from_segments(
        new_irreps, new_list, irreps_array.mul_ir, rotation_basis.shape[:-1]
    )


def subrepr_permutation(
    sub_f0: FrozenSet[int], perm_repr: FrozenSet[Tuple[int, Tuple[int, ...]]]
) -> FrozenSet[Tuple[int, Tuple[int, ...]]]:
    sor = sorted(sub_f0)
    return frozenset(
        {
            (s, tuple(sor.index(i) for i in p if i in sub_f0))
            for s, p in perm_repr
            if all(i in sub_f0 or i == j for j, i in enumerate(p))
        }
    )


def reduce_subgroup_permutation(
    sub_f0: FrozenSet[int],
    perm_repr: FrozenSet[Tuple[int, Tuple[int, ...]]],
    dims: Tuple[int, ...],
    return_dim: bool = False,
) -> np.ndarray:
    sub_perm_repr = subrepr_permutation(sub_f0, perm_repr)
    sub_dims = tuple(dims[i] for i in sub_f0)
    if len(sub_perm_repr) == 1:
        if return_dim:
            return prod(sub_dims)
        return np.eye(prod(sub_dims)).reshape((prod(sub_dims),) + sub_dims)
    base = reduce_permutation_base(sub_perm_repr, sub_dims)
    if return_dim:
        return len(base)
    permutation_basis = reduce_permutation_matrix(base, sub_dims)
    return np.reshape(
        permutation_basis,
        (-1,) + tuple(dims[i] if i in sub_f0 else 1 for i in range(len(dims))),
    )


@functools.lru_cache(maxsize=None)
def full_base_fn(dims: Tuple[int, ...]) -> List[Tuple[int, ...]]:
    return list(itertools.product(*(range(d) for d in dims)))


@functools.lru_cache(maxsize=None)
def reduce_permutation_base(
    perm_repr: FrozenSet[Tuple[int, Tuple[int, ...]]], dims: Tuple[int, ...]
) -> FrozenSet[FrozenSet[FrozenSet[Tuple[int, Tuple[int, ...]]]]]:
    full_base = full_base_fn(dims)  # (0, 0, 0), (0, 0, 1), (0, 0, 2), ... (3, 3, 3)
    # len(full_base) degrees of freedom in an unconstrained tensor

    # but there is constraints given by the group `formulas`
    # For instance if `ij=-ji`, then 00=-00, 01=-01 and so on
    base = set()
    for x in full_base:
        # T[x] is a coefficient of the tensor T and is related to other coefficient T[y]
        # if x and y are related by a formula
        xs = {(s, tuple(x[i] for i in p)) for s, p in perm_repr}
        # s * T[x] are all equal for all (s, x) in xs
        # if T[x] = -T[x] it is then equal to 0 and we lose this degree of freedom
        if (-1, x) not in xs:
            # the sign is arbitrary, put both possibilities
            base.add(frozenset({frozenset(xs), frozenset({(-s, x) for s, x in xs})}))

    # len(base) is the number of degrees of freedom in the tensor.

    return frozenset(base)


@functools.lru_cache(maxsize=None)
def reduce_permutation_matrix(
    base: FrozenSet[FrozenSet[FrozenSet[Tuple[int, Tuple[int, ...]]]]],
    dims: Tuple[int, ...],
) -> np.ndarray:
    base = sorted(
        [sorted([sorted(xs) for xs in x]) for x in base]
    )  # requested for python 3.7 but not for 3.8 (probably a bug in 3.7)

    # First we compute the change of basis (projection) between full_base and base
    d_sym = len(base)
    Q = np.zeros((d_sym, prod(dims)), np.float64)

    for i, x in enumerate(base):
        x = max(x, key=lambda xs: sum(s for s, x in xs))
        for s, e in x:
            j = 0
            for k, d in zip(e, dims):
                j *= d
                j += k
            Q[i, j] = s / len(x) ** 0.5

    np.testing.assert_allclose(Q @ Q.T, np.eye(d_sym))

    return Q.reshape(d_sym, *dims)


def _rounding(x: irreps_array.NumpyIrrepsArray) -> irreps_array.NumpyIrrepsArray:
    return irreps_array.NumpyIrrepsArray(
        x.irreps, round_to_sqrt_rational(x.array), x.layout
    )


def _sparsify(input: irreps_array.NumpyIrrepsArray) -> irreps_array.NumpyIrrepsArray:
    assert input.layout == irreps_array.mul_ir
    input = input.regroup()

    segments = []
    for x in input.segments:
        x = np.moveaxis(x, -2, 0)  # (mul, ...)
        x, _, _ = sparsify_matrix(x, round_fn=round_to_sqrt_rational)
        x = np.moveaxis(x, 0, -2)
        segments.append(x)

    return cue.from_segments(
        input.irreps, segments, irreps_array.mul_ir, input.shape[:-1]
    )



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools

import numpy as np
import pytest

import cuequivariance as cue


@pytest.mark.parametrize("Group", [cue.SU2, cue.SO3, cue.O3])
def test_rep(Group):
    # [X_i, X_j] = A_ijk X_k

    for r in itertools.islice(Group.iterator(), 6):
        A = r.A
        assert A.shape[0] == r.lie_dim
        X = r.X
        H = r.H
        assert r.dim == X.shape[1]
        assert r.dim == H.shape[1]


@pytest.mark.parametrize("Group", [cue.SU2, cue.SO3, cue.O3])
def test_is_scalar(Group):
    for r in itertools.islice(Group.iterator(), 6):
        assert r.is_scalar() == cue.Rep.is_scalar(r)


@pytest.mark.parametrize("Group", [cue.SU2, cue.SO3, cue.O3])
def test_dim(Group):
    for r in itertools.islice(Group.iterator(), 6):
        assert r.dim == cue.Rep.dim.fget(r)


def test_is_scalar_O3():
    assert cue.O3(0, 1).is_scalar()
    assert not cue.O3(0, -1).is_scalar()

    assert cue.O3(0, 1).is_trivial()
    assert not cue.O3(0, -1).is_trivial()

    r = cue.O3(0, 1)
    assert cue.Rep.is_scalar(r)
    assert cue.Rep.is_trivial(r)
    r = cue.O3(0, -1)
    assert not cue.Rep.is_scalar(r)
    assert not cue.Rep.is_trivial(r)


def test_exp_map():
    r = cue.SO3.from_string("1")
    np.testing.assert_allclose(
        r.exp_map(np.array([0.0, 0.0, 0.0]), np.array([])), np.eye(3)
    )
    np.testing.assert_allclose(
        r.exp_map(np.array([0.0, np.pi, 0.0]), np.array([])),
        np.diag([-1.0, 1.0, -1.0]),
        rtol=1e-15,
        atol=1e-15,
    )
    np.testing.assert_allclose(
        r.exp_map(np.array([2 * np.pi, 0.0, 0.0]), np.array([])),
        np.eye(3),
        rtol=1e-15,
        atol=1e-15,
    )
    r = cue.O3.from_string("1o")
    np.testing.assert_allclose(
        r.exp_map(np.array([0.0, 0.0, 0.0]), np.array([1])), -np.eye(3)
    )



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import pytest

import cuequivariance as cue


def test_rep_collection_context():
    explicit = cue.Irreps("O3", "0e + 1o + 2e")
    with cue.assume("O3"):
        implicit = cue.Irreps("0e + 1o + 2e")

    assert explicit == implicit

    with pytest.raises(ValueError):
        cue.Irreps("0e + 1o + 2e")


def test_layout_context():
    explicit = cue.NumpyIrrepsArray(cue.Irreps("O3", "0e + 1o"), np.ones(4), cue.ir_mul)

    with cue.assume(cue.ir_mul):
        implicit = cue.NumpyIrrepsArray(cue.Irreps("O3", "0e + 1o"), np.ones(4))

    assert explicit == implicit

    with pytest.raises(ValueError):
        cue.NumpyIrrepsArray(cue.Irreps("O3", "0e + 1o"), np.ones(4))


def test_rep_collection_and_layout_context():
    explicit = cue.NumpyIrrepsArray(cue.Irreps("O3", "0e + 1o"), np.ones(4), cue.ir_mul)

    with cue.assume("O3", cue.ir_mul):
        implicit = cue.NumpyIrrepsArray("0e + 1o", np.ones(4))

    assert explicit == implicit


def test_decorator():
    @cue.assume("O3", cue.ir_mul)
    def func():
        assert cue.get_irrep_scope() == cue.O3
        assert cue.get_layout_scope() == cue.ir_mul

    assert cue.get_irrep_scope(False) is None
    assert cue.get_layout_scope(False) is None
    func()
    assert cue.get_irrep_scope(False) is None
    assert cue.get_layout_scope(False) is None



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import cuequivariance.segmented_polynomials as sp
import numpy as np
from cuequivariance.group_theory import descriptors

import cuequivariance as cue


def test_dot1():
    d1 = cue.SegmentedTensorProduct.from_subscripts("iab_jb_ak+ijk")
    d1.add_path(None, None, None, c=np.random.randn(2, 2, 2), dims={"a": 2, "b": 3})
    d1.add_path(None, None, None, c=np.random.randn(2, 3, 2), dims={"a": 4, "b": 3})
    d1.add_path(0, 1, 0, c=np.random.randn(2, 3, 2))

    d2 = cue.SegmentedTensorProduct.from_subscripts("jb_b_+j")
    d2.add_path(None, None, None, c=np.random.randn(2), dims={"b": 3})
    d2.add_path(None, 0, None, c=np.random.randn(3))

    d3 = sp.dot(d1, d2, (1, 0))
    assert d3.subscripts == "iab,ak,b,+ijk"

    x0, x2 = np.random.randn(d1.operands[0].size), np.random.randn(d1.operands[2].size)
    y1 = np.random.randn(d2.operands[1].size)

    tmp = sp.compute_last_operand(d1.move_operand_last(1), x0, x2)
    z0 = sp.compute_last_operand(d2, tmp, y1)

    z1 = sp.compute_last_operand(d3, x0, x2, y1)

    np.testing.assert_allclose(z0, z1)


def make_examples():
    irreps_middle = cue.Irreps("SO3", "2x0 + 3x1")
    dx = descriptors.fully_connected_tensor_product(
        cue.Irreps("SO3", "4x0 + 3x1"),
        cue.Irreps("SO3", "3x0 + 5x1"),
        irreps_middle,
    ).polynomial.operations[0][1]
    assert dx.subscripts == "uvw,iu,jv,kw+ijk"
    dy = descriptors.channelwise_tensor_product(
        irreps_middle, cue.Irreps("SO3", "0 + 1 + 2"), cue.Irreps("SO3", "0 + 1")
    ).polynomial.operations[0][1]
    dy = dy.squeeze_modes("v")
    assert dy.subscripts == "u,iu,j,ku+ijk"
    dy = dy.add_or_rename_modes("w_kw_l_mw+klm")
    return dx, dy


def test_dot2():
    dx, dy = make_examples()

    dxy = sp.dot(dx, dy, (3, 1))
    assert dxy.subscripts == "uvw,iu,jv,w,l,mw+ijklm"

    x0, x1, x2 = [np.random.randn(dx.operands[i].size) for i in range(3)]
    y0, y2 = [np.random.randn(dy.operands[i].size) for i in [0, 2]]

    tmp = sp.compute_last_operand(dx, x0, x1, x2)
    A = sp.compute_last_operand(dy, y0, tmp, y2)

    B = sp.compute_last_operand(dxy, x0, x1, x2, y0, y2)

    np.testing.assert_allclose(A, B)


def test_trace():
    dx, dy = make_examples()

    d1 = sp.dot(dx, dy, (3, 1))
    d1 = d1.canonicalize_subscripts()
    d1 = d1.sort_paths()

    assert dy.subscripts == "w,kw,l,mw+klm"
    dy = dy.add_or_rename_modes("a_xa_y_za+xyz")
    d2 = sp.trace(sp.dot(dx, dy), (3, 4 + 1))
    d2 = d2.canonicalize_subscripts()
    d2 = d2.sort_paths()

    assert d1.subscripts == d2.subscripts
    assert d1.operands == d2.operands
    np.testing.assert_allclose(d1.indices, d2.indices)
    assert len(d1.paths) == len(d2.paths)

    for path1, path2 in zip(d1.paths, d2.paths):
        assert path1.indices == path2.indices
        np.testing.assert_allclose(path1.coefficients, path2.coefficients)



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .context_irrep_class import get_irrep_scope
from .irreps import MulIrrep, Irreps
from .irreps_layout import IrrepsLayout, mul_ir, ir_mul
from .irreps_and_layout import IrrepsAndLayout

from .context_layout import get_layout_scope
from .context_decorator import assume

from .numpy_irreps_array import NumpyIrrepsArray, from_segments, concatenate
from .reduced_tensor_product import (
    reduced_tensor_product_basis,
    reduced_symmetric_tensor_product_basis,
    reduced_antisymmetric_tensor_product_basis,
)

__all__ = [
    "get_irrep_scope",
    "MulIrrep",
    "Irreps",
    "IrrepsLayout",
    "mul_ir",
    "ir_mul",
    "IrrepsAndLayout",
    "get_layout_scope",
    "assume",
    "NumpyIrrepsArray",
    "from_segments",
    "concatenate",
    "reduced_tensor_product_basis",
    "reduced_symmetric_tensor_product_basis",
    "reduced_antisymmetric_tensor_product_basis",
]



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import cuequivariance as cue


def gatr_linear(mul_in: int, mul_out: int) -> cue.SegmentedTensorProduct:
    """
    subsrcipts: ``weights[uv],input[iu],output[iv]``

    References:
      - https://arxiv.org/pdf/2305.18415
      - `Source code <https://github.com/Qualcomm-AI-research/geometric-algebra-transformer/blob/3f967c978445648ef83d87190d32176f7fd91565/gatr/primitives/linear.py#L33-L43>`_

    Examples:
        >>> d = gatr_linear(32, 32)
        >>> d
        uv,iu,iv sizes=9216,512,512 num_segments=9,8,8 num_paths=12 i={1, 3} u=32 v=32
    """
    d = cue.SegmentedTensorProduct.from_subscripts("uv,iu,iv")

    for ope, mul in [(1, mul_in), (2, mul_out)]:
        one = d.add_segment(ope, (1, mul))
        e0 = d.add_segment(ope, (1, mul))
        ei = d.add_segment(ope, (3, mul))
        e0i = d.add_segment(ope, (3, mul))
        eij = d.add_segment(ope, (3, mul))
        e0ij = d.add_segment(ope, (3, mul))
        e123 = d.add_segment(ope, (1, mul))
        e0123 = d.add_segment(ope, (1, mul))

    for xs in [[one], [e0, ei], [e0i, eij], [e0ij, e123], [e0123]]:
        x, xs = xs[0], xs[1:]
        d.add_path(None, x, x, c=1)
        for y in xs:
            d.add_path(-1, y, y, c=1)

    d.add_path(None, one, e0, c=1)
    d.add_path(None, ei, e0i, c=1)
    d.add_path(None, eij, e0ij, c=1)
    d.add_path(None, e123, e0123, c=1)
    return d


def gatr_geometric_product() -> cue.SegmentedTensorProduct:
    """
    subsrcipts: ``input1[i],input2[j],output[k]+ijk``

    References:
      - https://arxiv.org/pdf/2305.18415
      - `Source code <https://github.com/Qualcomm-AI-research/geometric-algebra-transformer/blob/3f967c978445648ef83d87190d32176f7fd91565/gatr/primitives/bilinear.py#L63>`_

    Examples:
        >>> d = gatr_geometric_product(); d
        i,j,k+ijk sizes=16,16,16 num_segments=8,8,8 num_paths=60 i={1, 3} j={1, 3} k={1, 3}

        >>> d = d.append_modes_to_all_operands("u", dict(u=32)); d
        iu,ju,ku+ijk sizes=512,512,512 num_segments=8,8,8 num_paths=60 i={1, 3} j={1, 3} k={1, 3} u=32

        >>> d = d.normalize_paths_for_operand(2); d.paths[-1]
        op0[7]*op1[6]*op2[1]*[[[-0.25]]]

        >>> d = d.flatten_coefficient_modes(); d
        u,u,u sizes=512,512,512 num_segments=16,16,16 num_paths=192 u=32
    """
    return cue.SegmentedTensorProduct.from_base64(
        """
    eJztV9FuozAQ/BXE6+ETNsaW7leiPPRSekerS6OmfYry72eWcTcxNimlTdWm0gpGZne8s7uYZJdvn35v
    Vw/t5nGb/8rytrgt7n60t3d5keX3m+bhan3dPVjs8m3z51+zJrfFQi6LrL9UkQueLvdFdt44d883V49/
    u6Bd3q6v21XTE5RF1lnn6O7SGaByBlg5A9TOAGtngMYZoHVGuxKvBFRgkMRQAxr4KvJVgBK+isJKhhqw
    Am8Pa0DNDBrpKEqyYuh3M0xmwVBRDhWgwhY99A6aHTTIKiIjB00MGlDyqq9kDw2gL2oPLaCGYs2l1pS6
    ZOh5DfNakNWUQw2oeFVhi5p4JUPva+BriMEASoQZboDhqhuuuuFSGyIrAS12s8RrAX1RLRfVUpgbVjeq
    q/vm5qZdtc9j7gb9Z7nsng2g48WFJrDDMlzwq58t7CT8sHQOGGjtmIOXQhaRihWpYBHk0Ofnb96ZTpT4
    8sFT0JRhBU6UhKJE3Cm+IBPajpyEjPEcFGpEqxgXK8bUJpQEHZgiVxzrTc9SSu3ksJkdmVaO5xzDcRpr
    SKR/B1qnjcTMfkwZ9ZhWEYoNtcaGNRy+1CsRETeITc1XbAzmn1EXdMpc3Hv3ylGcN02vDR229UTlA6p4
    6wc08fYkaM53Kn5dre/4bT+pZDhssaK94GfaCzp0iWqDlKZ9tGYFfx803y/fm0/UUdWiH1K/80hy8VwG
    5YlIf7PAaZmG/zmH8Ow1QErL/f4/i15xrQ==
        """
    )


def gatr_outer_product() -> cue.SegmentedTensorProduct:
    """
    subsrcipts: ``input1[i],input2[j],output[k]+ijk``

    References:
      - https://arxiv.org/pdf/2305.18415
      - `Source code <https://github.com/Qualcomm-AI-research/geometric-algebra-transformer/blob/3f967c978445648ef83d87190d32176f7fd91565/gatr/primitives/bilinear.py#L88>`_

    Examples:
        >>> d = gatr_outer_product(); d
        i,j,k+ijk sizes=16,16,16 num_segments=8,8,8 num_paths=30 i={1, 3} j={1, 3} k={1, 3}
    """
    return cue.SegmentedTensorProduct.from_base64(
        """
    eJzVVMtugzAQ/BXka3GFbR5SfwVxSAlpnagJCukJ8e+1N0MKBjeqSBtFWpnReGZYLxYtaz5fm/Ko61PD
    XgKmw224e9LbHQsDdqir42q/tht5y5rq7aPakyzPRREG50XNLNgtujD4X595snp1eremlun9WpfVOSAK
    A1tWaJ7CFKA0BahMAcamABNTgKkpwMwUvZVyBaBEgqCEBDCFVpJWAgpoJdliQAWbpIQUMEGCogQFKKFV
    pCVBTIIYUEAQkzYFVNAmpE0AJdiU2BRQgM2IzQo74PJQbTa61JePYz7Pc1TYvQk0Niw0N4uFS/Tso9mu
    wru1M0ggbpzxTbkp3OflPjMfuCO39StnIRefF80TwtPUSEQtTXJ8PXre7xz4N01yMT/2yUl+6BF7GOIl
    rW+43+57G8i9OdwNcnO4N2jBjXqQO3GTeS8Y07IJ/8FtucsUbtjC8I82pouu+wK/DgJI
        """
    )



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
from cuequivariance.etc import linalg


def test_as_approx_integer_ratio():
    big = 1 << 62
    toobig = 1 << 63

    n, d = linalg.as_approx_integer_ratio(
        np.array([-0.5, 0.0, 0.1, 1.0, 10.0, big, toobig, 1 / toobig])
    )
    np.testing.assert_equal(
        n, [-1, 0, 3602879701896397, 1, 36028797018963968, big, 0, 0]
    )
    np.testing.assert_equal(d, [2, 1, 36028797018963968, 1, 3602879701896397, 1, 0, 0])


def test_round_to_rational():
    x0 = np.array([-12.0, -0.5, 0.0, 0.5, 1e10, 1e16, 0.75 + 1e-13])
    x1 = np.array([-12.0, -0.5, 0.0, 0.5, 1e10, 1e16, 0.75])
    np.testing.assert_equal(linalg.round_to_rational(x0), x1)


def test_round_to_sqrt_rational():
    x0 = np.array([0.0, -0.5, np.sqrt(3) + 1e-13])
    x1 = np.array([0.0, -0.5, np.sqrt(3)])
    np.testing.assert_equal(linalg.round_to_sqrt_rational(x0), x1)


def test_limit_denominator():
    np.testing.assert_equal(linalg.limit_denominator(1, 7, 7), [1, 7])
    np.testing.assert_equal(linalg.limit_denominator(4, 7 * 5, 7), [1, 7])



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Union

from cuequivariance.group_theory.irreps_array import IrrepsLayout

_layout: Union[None, IrrepsLayout] = None


def get_layout_scope(raising: bool = True) -> IrrepsLayout:
    if raising and _layout is None:
        raise ValueError(
            "No layout set in the context. Please specify the layout explicitly or use ``with cue.assume(layout):``."
        )

    return _layout


def push_layout_scope(layout):
    global _layout
    old_layout = _layout
    _layout = layout
    return old_layout


def pop_layout_scope(old_layout):
    global _layout
    _layout = old_layout



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Type, Union

import cuequivariance as cue
from cuequivariance.group_theory.representations import Irrep

_irrep_class: Union[None, str, Type[Irrep]] = None


def get_irrep_scope(raising: bool = True) -> Type[Irrep]:
    if raising and _irrep_class is None:
        raise ValueError(
            "No irrep class set in the context. Please specify the irrep class explicitly or use ``with cue.assume(irrep):``."
        )

    if isinstance(_irrep_class, str):
        return getattr(cue, _irrep_class)
    return _irrep_class


def push_irrep_scope(irrep_class):
    global _irrep_class
    old_irrep_class = _irrep_class
    _irrep_class = irrep_class
    return old_irrep_class


def pop_irrep_scope(old_irrep_class):
    global _irrep_class
    _irrep_class = old_irrep_class



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
from dataclasses import dataclass
from functools import wraps
from typing import Callable, Set

import networkx
import numpy as np

# With a mantissa of 53 bits, the maximum denominator is 2**53
# We use 2**40 to give some room for the numerical errors
DEFAULT_MAX_DENOMINATOR = 2**40


def normalize_integer_ratio(n, d):
    assert np.all(d > 0)
    g = np.gcd(n, d)
    g = np.where(d < 0, -g, g)
    return n // g, d // g


def _as_approx_integer_ratio(x):
    # only for 0 < x <= 1
    assert x.dtype == np.float64
    assert np.all(0 < x)
    assert np.all(x <= 1)

    big = 1 << 62
    n = np.floor(x * big).astype(np.int64)
    d = np.round(n / x).astype(np.int64)

    # if x is too small, then n = 0 and d = 0
    return n, d


def as_approx_integer_ratio(x: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    # return (0, 0) if x is too small or too large
    x = np.asarray(x, dtype=np.float64)
    sign = np.sign(x).astype(np.int64)
    x = np.abs(x)

    # case: x = 0
    n, d = np.zeros_like(x, np.int64), np.ones_like(x, np.int64)

    # case: 0 < x <= 1
    mask = (x > 0.0) & (x <= 1.0)
    n, d = np.where(
        mask,
        _as_approx_integer_ratio(np.where(mask, x, 1.0)),
        (n, d),
    )

    # case: x > 1
    mask = x > 1.0
    n, d = np.where(
        mask,
        _as_approx_integer_ratio(1 / np.where(mask, x, 1.0))[::-1],
        (n, d),
    )

    mask = d > 0
    n, d = np.where(
        mask, normalize_integer_ratio(sign * n, np.where(mask, d, 1)), (n, d)
    )

    return n, d


def limit_denominator(n, d, max_denominator: int):
    # (n, d) = must be normalized
    n = np.asarray(n, dtype=np.int64)
    d = np.asarray(d, dtype=np.int64)
    assert np.all(d > 0)
    assert np.all(np.gcd(n, d) == 1)

    n0, d0 = n, d
    p0, q0, p1, q1 = (
        np.zeros_like(n),
        np.ones_like(n),
        np.ones_like(n),
        np.zeros_like(n),
    )
    while True:
        a = n // d
        q2 = q0 + a * q1
        stop = (q2 > max_denominator) | (d0 <= max_denominator)
        if np.all(stop):
            break
        p0, q0, p1, q1 = np.where(stop, (p0, q0, p1, q1), (p1, q1, p0 + a * p1, q2))
        n, d = np.where(stop, (n, d), (d, n - a * d))

    with np.errstate(divide="ignore"):
        k = (max_denominator - q0) // q1
    n1, d1 = p0 + k * p1, q0 + k * q1
    n2, d2 = p1, q1
    with np.errstate(over="ignore"):
        mask = np.abs(d1 * (n2 * d0 - n0 * d2)) <= np.abs(d2 * (n1 * d0 - n0 * d1))
    return np.where(
        d0 <= max_denominator,
        (n0, d0),
        np.where(mask, (n2, d2), (n1, d1)),
    )


def support_complex(f: Callable) -> Callable:
    @wraps(f)
    def wrapper(x, *args, **kwargs):
        if np.iscomplex(x).any():
            r = f(np.real(x), *args, **kwargs)
            i = f(np.imag(x), *args, **kwargs)
            return r + 1j * i
        return f(x, *args, **kwargs)

    return wrapper


@support_complex
def round_to_rational(
    x: np.ndarray, max_denominator: int = DEFAULT_MAX_DENOMINATOR
) -> np.ndarray:
    """Round a number to the closest number of the form ``n/d`` for ``d <= max_denominator``"""
    x = np.asarray(x, dtype=np.float64)
    n, d = as_approx_integer_ratio(x)

    mask = d > 0
    n, d = limit_denominator(n, np.where(mask, d, 1), max_denominator)
    return np.where(mask, n / d, 0.0)


@support_complex
def round_to_sqrt_rational(
    x: np.ndarray, max_denominator: int = DEFAULT_MAX_DENOMINATOR
) -> np.ndarray:
    """Round a number to the closest number of the form ``sqrt(n/d)`` for ``d <= max_denominator``"""
    x = np.asarray(x, dtype=np.float64)
    sign = np.sign(x)
    n, d = as_approx_integer_ratio(x**2)

    mask = d > 0
    n, d = limit_denominator(n, np.where(mask, d, 1), max_denominator)
    return np.where(mask, sign * np.sqrt(n / d), 0.0)


def gram_schmidt(A: np.ndarray, *, epsilon=1e-5, round_fn=lambda x: x) -> np.ndarray:
    """
    Orthogonalize a matrix using the Gram-Schmidt process.
    """
    assert A.ndim == 2, "Gram-Schmidt process only works for matrices."
    assert A.dtype in [
        np.float64,
        np.complex128,
    ], "Gram-Schmidt process only works for float64 matrices."
    Q = []
    for i in range(A.shape[0]):
        v = A[i]
        for w in Q:
            v -= np.dot(np.conj(w), v) * w
        norm = np.linalg.norm(v)
        if norm > epsilon:
            v = round_fn(v / norm)
            Q += [v]
    return np.stack(Q) if len(Q) > 0 else np.empty((0, A.shape[1]))


def basis_intersection(
    basis1: np.ndarray, basis2: np.ndarray, *, epsilon=1e-5, round_fn=lambda x: x
) -> tuple[np.ndarray, np.ndarray]:
    """Compute the intersection of two bases

    Args:
        basis1 (np.ndarray): A basis, shape ``(n1, d)``
        basis2 (np.ndarray): Another basis, shape ``(n2, d)``
        epsilon (float, optional): Tolerance for the norm of the vectors. Defaults to 1e-4.
        round_fn (function, optional): Function to round the vectors. Defaults to lambda x: x.

    Returns:
        (tuple): tuple containing:

            np.ndarray: A projection matrix that projects vectors of the first basis in the intersection of the two bases.
                Shape ``(dim_intersection, n1)``
            np.ndarray: A projection matrix that projects vectors of the second basis in the intersection of the two bases.
                Shape ``(dim_intersection, n2)``

    Examples:
        >>> basis1 = np.array([[1, 0, 0], [0, 0, 1.0]])
        >>> basis2 = np.array([[1, 1, 0], [0, 1, 0.0]])
        >>> P1, P2 = basis_intersection(basis1, basis2)
        >>> P1 @ basis1
        array([[1., 0., 0.]])
    """
    assert basis1.ndim == 2  # (n1, d)
    assert basis2.ndim == 2  # (n2, d)
    assert basis1.shape[1] == basis2.shape[1]

    p = np.concatenate(
        [
            np.concatenate([basis1 @ basis1.T, -basis1 @ basis2.T], axis=1),
            np.concatenate([-basis2 @ basis1.T, basis2 @ basis2.T], axis=1),
        ],
        axis=0,
    )
    p = round_fn(p)
    # p.shape = (n1 + n2, n1 + n2)

    w, v = np.linalg.eigh(p)
    v = v[:, w < epsilon]
    # ni = dim_intersection <= min(n1, n2)
    # v.shape = (n1 + n2, ni)

    x1 = v[: basis1.shape[0], :]  # (n1, ni)
    x1 = gram_schmidt(x1 @ x1.T, epsilon=epsilon, round_fn=round_fn)
    # x1.shape = (ni, n1)

    x2 = v[basis1.shape[0] :, :]  # (n2, ni)
    x2 = gram_schmidt(x2 @ x2.T, epsilon=epsilon, round_fn=round_fn)
    # x2.shape = (ni, n2)

    return x1, x2


TY_PERM = tuple[int, ...]


def perm_compose(p1: TY_PERM, p2: TY_PERM) -> TY_PERM:
    # p: i |-> p[i]

    # [p1.p2](i) = p1(p2(i)) = p1[p2[i]]
    return tuple(p1[p2[i]] for i in range(len(p1)))


def perm_inverse(p: TY_PERM) -> TY_PERM:
    return tuple(p.index(i) for i in range(len(p)))


def perm_to_cycles(p: TY_PERM) -> Set[tuple[int]]:
    n = len(p)

    cycles = set()

    for i in range(n):
        c = [i]
        while p[i] != c[0]:
            i = p[i]
            c += [i]
        if len(c) >= 2:
            i = c.index(min(c))
            c = c[i:] + c[:i]
            cycles.add(tuple(c))

    return cycles


def perm_sign(p: TY_PERM) -> int:
    s = 1
    for c in perm_to_cycles(p):
        if len(c) % 2 == 0:
            s = -s
    return s


def triu_coo(input: np.ndarray, ndim: int) -> tuple[np.ndarray, np.ndarray]:
    indices = np.nonzero(input)
    values = input[indices]
    indices = np.stack(indices, axis=1)
    indices = np.concatenate(
        [np.sort(indices[:, :ndim], axis=1), indices[:, ndim:]], axis=1
    )
    indices, inverse = np.unique(indices, axis=0, return_inverse=True)
    inverse = inverse.reshape(-1)
    values = np.bincount(inverse, values)
    return indices, values


def triu_array(input: np.ndarray, ndim: int) -> np.ndarray:
    indices, values = triu_coo(input, ndim)
    input = np.zeros_like(input)
    input[tuple(indices.T)] = values
    return input


@dataclass
class SparsifyResult:
    pass


@dataclass
class DisjointRows(SparsifyResult):
    pass


@dataclass
class ReplaceRow(SparsifyResult):
    a0: float
    a1: float
    row: np.ndarray
    which: int


@dataclass
class AlreadySparse(SparsifyResult):
    nc: int
    nx: int
    n0: int
    n1: int


def sparsify_rows(
    row0: np.ndarray, row1: np.ndarray, round_fn=lambda x: x
) -> SparsifyResult:
    """Try to find a sparse linear combination of row0 and row1.

    Args:
        row0: The first row.
        row1: The second row.

    Returns:
        The coefficients of the sparse linear combination.
        The linear combination.
        The index of the row that worth sparsifying.

    One could always reorder the rows so that:

        row0:
        [maximally comensurate][nonzero][nonzero][       ][     ]
        row1:
        [maximally comensurate][nonzero][       ][nonzero][     ]
             nc         nx       n0       n1
    """
    res = np.zeros_like(row0)

    i = np.where((row0 != 0) & (row1 != 0))[0]
    if len(i) == 0:
        return DisjointRows()

    r = row0[i] / row1[i]
    r = round_fn(r)
    r, c = np.unique(r, return_counts=True)
    r, nc = r[np.argmax(c)], np.max(c)

    n0 = np.sum((row0 != 0) & (row1 == 0))
    n1 = np.sum((row0 == 0) & (row1 != 0))

    if nc > min(n0, n1):
        res = row0 - r * row1
        j = 1 if n0 < n1 else 0
        v = np.linalg.norm(row0 if j == 0 else row1) / np.linalg.norm(res)
        res = res * v
        res = round_fn(res)
        return ReplaceRow(a0=v, a1=-r * v, row=res, which=j)

    return AlreadySparse(nc=nc, nx=len(i) - nc, n0=n0, n1=n1)


def sparsify_matrix(
    input: np.ndarray, max_iterations: int = 20, round_fn=lambda x: x
) -> tuple[np.ndarray, np.ndarray, networkx.Graph]:
    """Reduce the number of non-zero elements in a matrix.

    Args:
        input: The matrix to reduce.

    Returns:
        An invertible matrix q such that q @ input is sparser.
    """
    row_shape = input.shape[1:]
    input = np.reshape(input, (len(input), np.prod(row_shape)))
    assert input.ndim == 2

    non_zero_columns = np.where(np.any(input != 0, axis=0))[0]
    x = input[:, non_zero_columns]
    q = np.eye(x.shape[0])

    graph = networkx.Graph()
    graph.add_nodes_from(range(x.shape[0]))
    for i0, i1 in itertools.combinations(range(x.shape[0]), 2):
        graph.add_edge(i0, i1)

    iterations = 0
    hope = True
    while hope and iterations < max_iterations:
        hope = False

        next_graph = networkx.Graph()
        next_graph.add_nodes_from(range(x.shape[0]))

        for i0, i1 in graph.edges:
            result = sparsify_rows(x[i0], x[i1], round_fn)

            match result:
                case DisjointRows():
                    pass
                case ReplaceRow(a0=a0, a1=a1, row=row, which=which):
                    hope = True
                    which = i0 if which == 0 else i1
                    x[which] = row
                    q[which] = a0 * q[i0] + a1 * q[i1]

                    next_graph.add_edge(i0, i1)
                    for i in graph.neighbors(i0):
                        if i != i1:
                            next_graph.add_edge(i, i1)
                    for i in graph.neighbors(i1):
                        if i != i0:
                            next_graph.add_edge(i, i0)
                case AlreadySparse():
                    next_graph.add_edge(i0, i1)
                case _:
                    raise ValueError(f"Unknown result type: {result}")

        iterations += 1
        graph = next_graph

    output = np.zeros_like(input)
    output[:, non_zero_columns] = x
    output = np.reshape(output, (len(output),) + row_shape)
    return output, q, graph



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
import re
from typing import Any, Callable, NamedTuple, Optional, Sequence, Type, Union

import cuequivariance as cue


@dataclasses.dataclass(frozen=True)
class MulIrrep:
    mul: int
    ir: cue.Irrep

    def __repr__(self):
        if self.mul == 1:
            return f"{self.ir}"
        return f"{self.mul}x{self.ir}"

    def __iter__(self):
        return iter((self.mul, self.ir))


# This class is inspired from https://github.com/e3nn/e3nn-jax/blob/245e17eb23deaccad9f2c9cfd40fe40515e3c074/e3nn_jax/_src/irreps.py
@dataclasses.dataclass(init=False, frozen=True)
class Irreps:
    """
    Direct sum of irreducible representations with multiplicities.

    For more information, see :ref:`tuto_irreps`.

    Args:
        irrep_class: Class of the irreducible representations
            (e.g., :class:`SU2 <cuequivariance.SU2>`, :class:`SO3 <cuequivariance.SO3>`, :class:`O3 <cuequivariance.O3>`).
        input: A description of multiplicities of each irreducible representation.

    Examples:
        >>> Irreps("SO3", "16x0 + 4x1")
        16x0+4x1

        >>> Irreps(cue.SO3, "16x0 + 4x1")
        16x0+4x1

        >>> Irreps(cue.SO3, [(16, 0), (4, 1)])
        16x0+4x1

        >>> with cue.assume("SO3"):
        ...     Irreps("16x0 + 4x1")
        16x0+4x1

    .. rubric:: Methods
    """

    irrep_class: Type[cue.Irrep]
    _mulirreps: tuple[MulIrrep, ...]

    def __init__(self, *args):
        if len(args) == 1 and isinstance(args[0], Irreps):
            args = (args[0].irrep_class, args[0]._mulirreps)

        if len(args) == 1 and isinstance(args[0], cue.Irrep):
            args = (type(args[0]), [(1, args[0])])

        if len(args) == 1:
            args = (cue.get_irrep_scope(), args[0])

        if len(args) != 2:
            raise ValueError(
                "Irreps requires two arguments: irrep_collection and input."
                f" Got {len(args)} arguments."
            )

        irrep_class: Any = args[0]
        if isinstance(irrep_class, str):
            irrep_class = getattr(cue, irrep_class)
        if isinstance(irrep_class, cue.Irreps):
            irrep_class = irrep_class.irrep_class
        if isinstance(irrep_class, cue.Irrep):
            irrep_class = type(irrep_class)
        assert isinstance(irrep_class, type) and issubclass(irrep_class, cue.Irrep)

        irrep_class: Type[cue.Irrep]
        input: Union[Irreps, str, Sequence[MulIrrep, cue.Irrep]] = args[1]

        mulreps = []

        if isinstance(input, str):
            for mul_rep_str in [] if input.strip() == "" else input.split("+"):
                if "x" in mul_rep_str:
                    mul, rep_str = mul_rep_str.split("x")
                    mul = int(mul)
                    rep_str = rep_str.strip()
                else:
                    mul = 1
                    rep_str = mul_rep_str.strip()

                if not re.match(irrep_class.regexp_pattern(), rep_str):
                    raise ValueError(
                        f"Invalid representation string: {rep_str}, expected pattern: {irrep_class.regexp_pattern().pattern}"
                    )

                rep = irrep_class.from_string(rep_str)
                mulreps.append(MulIrrep(mul=mul, ir=rep))

        elif isinstance(input, Irreps):
            if irrep_class != input.irrep_class:
                raise ValueError(
                    f"Invalid representation: {input.irrep_class}, expected {irrep_class}"
                )
            mulreps = input._mulirreps
        else:
            for x in input:
                if isinstance(x, MulIrrep):
                    if not isinstance(x.ir, irrep_class):
                        raise ValueError(
                            f"Invalid representation: {x.ir}, expected subclass of {irrep_class}"
                        )
                    mulreps.append(x)
                elif isinstance(x, irrep_class):
                    mulreps.append(MulIrrep(mul=1, ir=x))
                else:
                    try:
                        mul, ir = x
                    except ValueError:
                        raise ValueError(
                            f"Invalid representation: {x}, expected (mul, irrep)"
                        )

                    if not isinstance(mul, int):
                        raise ValueError(f"Invalid multiplicity: {mul}, expected int")

                    try:
                        ir = irrep_class._from(ir)
                    except ValueError:
                        raise ValueError(
                            f"Invalid representation: {ir}, expected {irrep_class}"
                        )

                    mulreps.append(MulIrrep(mul=mul, ir=ir))

        object.__setattr__(self, "irrep_class", irrep_class)
        object.__setattr__(self, "_mulirreps", tuple(mulreps))

    def __len__(self):
        return len(self._mulirreps)

    def __iter__(self):
        return iter(self._mulirreps)

    def __getitem__(self, index: Union[int, slice]) -> Union[MulIrrep, Irreps]:
        x = self._mulirreps[index]

        if isinstance(index, slice):
            return Irreps(self.irrep_class, x)
        return x

    def __contains__(self, rep: Union[str, cue.Irrep]) -> bool:
        # This function does not check the multiplicity of the representation!
        rep = self.irrep_class._from(rep)
        return any(mulrep.ir == rep for mulrep in self)

    def __repr__(self):
        return "+".join(f"{mulrep}" for mulrep in self)

    def new_scalars(self, mul: int) -> Irreps:
        """
        Return a representation with all scalar representations.

        Examples:
            >>> Irreps("SO3", "32x1").new_scalars(2)
            2x0
        """
        return Irreps(self.irrep_class, [(mul, self.irrep_class.trivial())])

    def count(self, rep: Union[str, cue.Irrep]) -> int:
        """
        Count the total multiplicity of a representation.

        Examples:
            >>> Irreps("SO3", "100x0 + 20x1 + 10x0").count("0")
            110
        """
        rep = self.irrep_class._from(rep)
        return sum(mul for mul, r in self if r == rep)

    @property
    def dim(self) -> int:
        """
        Total dimension of the representation.

        Examples:
            >>> Irreps("SO3", "100x0 + 10x1").dim
            130
        """
        return sum(mul * rep.dim for mul, rep in self)

    @property
    def num_irreps(self) -> int:
        """
        Return the number of irreducible representations.

        Examples:
            >>> Irreps("SO3", "100x0 + 10x1").num_irreps
            110
        """
        return sum(mul for mul, _ in self)

    @property
    def muls(self) -> list[int]:
        """
        List of multiplicities.

        Examples:
            >>> Irreps("SO3", "100x0 + 10x1").muls
            [100, 10]
        """
        return [mul for mul, _ in self]

    def slices(self) -> list[slice]:
        """
        List of slices for each segment of the representation.

        Examples:
            >>> Irreps("SO3", "100x0 + 10x1").slices()
            [slice(0, 100, None), slice(100, 130, None)]
        """
        s = []
        i = 0
        for mul, rep in self:
            d = mul * rep.dim
            s.append(slice(i, i + d))
            i += d
        return s

    def is_scalar(self) -> bool:
        """
        All representations are scalar.

        Note:
            This function does not check the multiplicity of the representation.

        Examples:
            >>> Irreps("SO3", "100x0 + 0x1").is_scalar()
            False

            >>> Irreps("SO3", "100x0").is_scalar()
            True
        """
        # Does not check the multiplicity of the representation!
        return all(rep.is_scalar() for mul, rep in self)

    def __add__(self, other):
        other = Irreps(self.irrep_class, other)
        return Irreps(self.irrep_class, self._mulirreps + other._mulirreps)

    def __radd__(self, other):
        return Irreps(self.irrep_class, other) + self

    def __mul__(self, other):
        if isinstance(other, int):
            return Irreps(
                self.irrep_class, [MulIrrep(mul * other, rep) for mul, rep in self]
            )
        return NotImplemented  # pragma: no cover

    def __rmul__(self, other):
        return self * other

    def __floordiv__(self, other):
        if isinstance(other, int):
            return Irreps(
                self.irrep_class, [MulIrrep(mul // other, rep) for mul, rep in self]
            )
        return NotImplemented  # pragma: no cover

    def __eq__(self, other):
        try:
            other = Irreps(self.irrep_class, other)
        except ValueError:
            return False
        return (
            self.irrep_class == other.irrep_class
            and self._mulirreps == other._mulirreps
        )

    def merge_consecutive(self) -> Irreps:
        """
        Merge consecutive segments with the same representation.

        Examples:
            >>> Irreps("SO3", "1 + 1 + 0 + 1").merge_consecutive()
            2x1+0+1
        """
        out = []
        for mul, rep in self:
            if out and out[-1][1] == rep:
                out[-1] = (out[-1][0] + mul, rep)
            else:
                out.append((mul, rep))
        return Irreps(self.irrep_class, out)

    def remove_zero_multiplicities(self) -> Irreps:
        """
        Remove zero multiplicities.

        Examples:
            >>> Irreps("SO3", "1 + 0x2 + 1").remove_zero_multiplicities()
            1+1
        """
        return Irreps(self.irrep_class, [(mul, rep) for mul, rep in self if mul != 0])

    def simplify(self) -> Irreps:
        """
        Simplify the representation by removing zero multiplicities and merging consecutive tuples.

        Examples:
            >>> Irreps("SO3", "1 + 0x2 + 1").simplify()
            2x1
        """
        return self.remove_zero_multiplicities().merge_consecutive()

    def sort(self) -> SortResult:
        """
        Sort the representation.

        Returns:
            SortResult: The sorted representation and associated permutations.

        Examples:
            >>> Irreps("SO3", "1 + 2 + 0 + 1").sort()
            SortResult(irreps=0+1+1+2, perm=(1, 3, 0, 2), inv=(2, 0, 3, 1))
        """

        def inverse(p):
            return tuple(p.index(i) for i in range(len(p)))

        out = sorted([(rep, i, mul) for i, (mul, rep) in enumerate(self)])
        inv = tuple(i for rep, i, mul in out)
        perm = inverse(inv)
        irreps = Irreps(self.irrep_class, [(mul, rep) for rep, i, mul in out])
        return SortResult(irreps, perm, inv)

    def regroup(self) -> Irreps:
        """
        Regroup the representation by sorting and simplifying.

        Examples:
            >>> Irreps("SO3", "1 + 2 + 0 + 1").regroup()
            0+2x1+2
        """
        return self.sort().irreps.simplify()

    def set_mul(self, mul: int) -> Irreps:
        """
        Set the multiplicity of all segments.

        Examples:
            >>> Irreps("SO3", "3x0 + 2x0 + 4x1").set_mul(2)
            2x0+2x0+2x1
        """
        return Irreps(self.irrep_class, [(mul, rep) for _, rep in self])

    def filter(
        self,
        *,
        keep: Union[str, Sequence[cue.Irrep], Callable[[MulIrrep], bool], None] = None,
        drop: Union[str, Sequence[cue.Irrep], Callable[[MulIrrep], bool], None] = None,
        mask: Optional[Sequence[bool]] = None,
    ) -> Irreps:
        """
        Filter the representation.

        Args:
            keep (str, list of Irrep, callable, optional): Keep only the specified representations.
            drop (str, list of Irrep, callable, optional): Drop the specified representations.

        Examples:
            >>> Irreps("SO3", "4x0 + 4x1 + 2x2").filter(keep="0 + 1")
            4x0+4x1

            >>> Irreps("SO3", "4x0 + 4x1 + 2x2").filter(drop="0 + 1")
            2x2
        """
        if mask is None:
            mask = self.filter_mask(keep=keep, drop=drop)
        return Irreps(
            self.irrep_class, [mulrep for keep, mulrep in zip(mask, self) if keep]
        )

    def filter_mask(
        self,
        *,
        keep: Union[str, Sequence[cue.Irrep], Callable[[MulIrrep], bool], None] = None,
        drop: Union[str, Sequence[cue.Irrep], Callable[[MulIrrep], bool], None] = None,
    ) -> list[bool]:
        if keep is not None:
            if drop is not None:
                raise ValueError(
                    "Only one of `keep` or `drop` must be defined."
                )  # pragma: no cover
            else:
                return self._filter_keep(keep)
        else:
            if drop is not None:
                return self._filter_drop(drop)
            else:
                raise ValueError(
                    "One of `keep` or `drop` must be defined."
                )  # pragma: no cover

    def _filter_keep(
        self, keep: Union[str, Sequence[cue.Irrep], Callable[[MulIrrep], bool]]
    ):
        if callable(keep):
            return [keep(mulrep) for mulrep in self]

        from .irrep_utils import into_list_of_irrep

        keep = into_list_of_irrep(self.irrep_class, keep)

        if not all(isinstance(rep, cue.Irrep) for rep in keep):
            raise ValueError(f"Invalid `keep` {keep}, expected a list of Irrep")

        return [mulrep.ir in keep for mulrep in self]

    def _filter_drop(
        self, drop: Union[str, Sequence[cue.Irrep], Callable[[MulIrrep], bool]]
    ):
        if callable(drop):
            return [not drop(mulrep) for mulrep in self]

        from .irrep_utils import into_list_of_irrep

        drop = into_list_of_irrep(self.irrep_class, drop)

        if not all(isinstance(rep, cue.Irrep) for rep in drop):
            raise ValueError(f"Invalid `drop` {drop}, expected a list of Irrep")

        return [mulrep.ir not in drop for mulrep in self]

    def layout_insensitive(self) -> bool:
        """
        True if the representation is layout insensitive.

        Examples:
            >>> Irreps("SO3", "100x0 + 1x1 + 1x2").layout_insensitive()
            True

            >>> Irreps("SO3", "100x0 + 2x1").layout_insensitive()
            False
        """
        for mul, ir in self:
            if mul > 1 and ir.dim > 1:
                return False
        return True


class SortResult(NamedTuple):
    irreps: Irreps
    perm: tuple[int, ...]
    inv: tuple[int, ...]



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

import numpy as np

import cuequivariance as cue


# The function escn_iu_ju_ku below is a 1:1 adaptation of https://github.com/e3nn/e3nn-jax/blob/a2a81ab451b9cd597d7be27b3e1faba79457475d/e3nn_jax/experimental/linear_shtp.py#L38-L165
def escn_tp(
    irreps_in: cue.Irreps,
    irreps_out: cue.Irreps,
    m_max: Optional[int] = None,
    l_max: Optional[int] = None,
) -> cue.EquivariantPolynomial:
    """
    subsrcipts: ``weights[uv],input[u],output[v]``

    Tensor Product part of the eSCN convolution introduced in https://arxiv.org/pdf/2302.03655.pdf

    Args:
        irreps_in (Irreps): Irreps of the input.
        irreps_out (Irreps): Irreps of the output.
        m_max (int, optional): Maximum angular resolution around the principal axis.
        l_max (int, optional): Maximum angular resolution along the principal axis.

    Returns:
        EquivariantPolynomial:
            Descriptor of the tensor product part of the eSCN convolution.

            - Operand 0: weights
            - Operand 1: input
            - Operand 2: output
    """
    assert irreps_in.irrep_class == irreps_out.irrep_class
    G = irreps_in.irrep_class
    if G not in [cue.SO3, cue.O3]:
        # TODO: we could support SU2 since it shares the same Clebsch-Gordan coefficients as SO3 and O3
        raise NotImplementedError("Only SO3 and O3 are supported")

    if l_max is not None:

        def pr(mul_ir: cue.MulIrrep) -> bool:
            ir = mul_ir.ir

            return any(abs(other.l - ir.l) <= l_max for _, other in irreps_in)

        irreps_out = irreps_out.filter(keep=pr)

    d = cue.SegmentedTensorProduct.from_subscripts("iuv,ju,kv+ijk")

    for mul, ir in irreps_in:
        d.add_segment(1, (ir.dim, mul))
    for mul, ir in irreps_out:
        d.add_segment(2, (ir.dim, mul))

    for i1, (mul1, ir1) in enumerate(irreps_in):
        for i2, (mul2, ir2) in enumerate(irreps_out):
            ell = min(ir1.l, ir2.l)

            if l_max is not None:
                if abs(ir1.l - ir2.l) > l_max:
                    continue
            if m_max is not None:
                ell = min(ell, m_max)

            # Scaled rotation (2 degrees of freedom per |m|)
            c = np.zeros((2 * ell + 1, ir1.dim, ir2.dim))
            for m in range(-ell, ell + 1):
                # "cosine" part
                c[ell - abs(m), ir1.l + m, ir2.l + m] = 1.0

                # "sine" part
                if m != 0:
                    c[ell + abs(m), ir1.l - m, ir2.l + m] = 1.0 if m > 0 else -1.0

            if G == cue.SO3:
                pass  # keep all the degrees of freedom
            elif G == cue.O3:
                if (-1) ** (ir1.l + ir2.l) == ir1.p * ir2.p:
                    # Symmetric case: keep only the "cosine" part
                    c = c[: ell + 1]
                elif ell > 0:
                    # Antisymmetric case: keep only the "sine" part
                    c = c[ell + 1 :]
                else:
                    c = None

            if c is not None:
                d.add_path(None, i1, i2, c=c)

    d = d.normalize_paths_for_operand(2)
    d = d.flatten_coefficient_modes()
    return cue.EquivariantPolynomial(
        [
            cue.IrrepsAndLayout(irreps_in.new_scalars(d.operands[0].size), cue.ir_mul),
            cue.IrrepsAndLayout(irreps_in, cue.ir_mul),
        ],
        [cue.IrrepsAndLayout(irreps_out, cue.ir_mul)],
        cue.SegmentedPolynomial.eval_last_operand(d),
    )


def escn_tp_compact(
    irreps_in: cue.Irreps,
    irreps_out: cue.Irreps,
    m_max: Optional[int] = None,
) -> cue.SegmentedPolynomial:
    """
    subsrcipts: ``weights[uv],input[u],output[v]``

    Tensor Product part of the eSCN convolution introduced in https://arxiv.org/pdf/2302.03655.pdf

    This "compact" implementation puts the L index contiguous in memory.
    This allows to create bigger segments and less paths.

    Args:
        irreps_in (Irreps): Irreps of the input.
        irreps_out (Irreps): Irreps of the output.
        m_max (int, optional): Maximum angular resolution around the principal axis.

    Returns:
        SegmentedTensorProduct:
            Descriptor of the tensor product part of the eSCN convolution.

            - Operand 0: weights
            - Operand 1: input
            - Operand 2: output
    """
    assert irreps_in.irrep_class == irreps_out.irrep_class
    G = irreps_in.irrep_class
    if G not in [cue.SO3]:
        raise NotImplementedError("Only SO3 is supported")

    d = cue.SegmentedTensorProduct.from_subscripts("uv,u,v")

    l_max_in = max(ir.l for _, ir in irreps_in)
    for m in range(-l_max_in, l_max_in + 1):
        mulirs = [(mul, ir) for mul, ir in irreps_in if abs(m) <= ir.l]
        mul = sum(mul for mul, _ in mulirs)
        d.add_segment(1, (mul,))

    l_max_out = max(ir.l for _, ir in irreps_out)
    for m in range(-l_max_out, l_max_out + 1):
        mulirs = [(mul, ir) for mul, ir in irreps_out if abs(m) <= ir.l]
        mul = sum(mul for mul, _ in mulirs)
        d.add_segment(2, (mul,))

    if m_max is None:
        m_max = min(l_max_in, l_max_out)

    # m = 0
    d.add_path(None, l_max_in, l_max_out, c=1.0)

    for m in range(1, min(m_max, l_max_in, l_max_out) + 1):
        # "cosine" part
        d.add_path(None, l_max_in - m, l_max_out - m, c=1.0)
        i = d.operands[0].num_segments - 1
        d.add_path(i, l_max_in + m, l_max_out + m, c=1.0)

        # "sine" part
        d.add_path(None, l_max_in + m, l_max_out - m, c=1.0)
        i = d.operands[0].num_segments - 1
        d.add_path(i, l_max_in - m, l_max_out + m, c=-1.0)

    d = d.normalize_paths_for_operand(2)
    # TODO: return an EquivariantPolynomial using SphericalSignal
    return cue.SegmentedPolynomial.eval_last_operand(d)


class SphericalSignal(cue.Rep):
    def __init__(self, mul: int, l_max: int, m_max: int):
        self.mul = mul
        self.l_max = l_max
        self.m_max = m_max

    # TODO
